import logging, struct
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from enum import Enum
from io import BytesIO
from selectors import DefaultSelector, EVENT_READ
from socket import socket, AF_INET6, SOCK_DGRAM, SOCK_STREAM

@dataclass
class Header:
    class Opcode(Enum):
        QUERY = 0
        IQUERY = 1
        STATUS = 2
        NOTIFY = 4
        UPDATE = 5

    class Rcode(Enum):
        NO_ERROR = 0
        FORMAT_ERROR = 1
        SERVER_FAILURE = 2
        NAME_ERROR = 3
        NOT_IMPLEMENTED = 4
        REFUSED = 5
        YX_DOMAIN = 6
        YX_RR_SET = 7
        NX_RR_SET = 8
        NOT_AUTH = 9
        NOT_ZONE = 10

    transid: int
    isresp: bool
    opcode: Opcode
    aa: bool
    tc: bool
    rd: bool
    ra: bool
    rcode: Rcode
    qdcount: int
    ancount: int
    nscount: int
    arcount: int

    @classmethod
    def decode(cls, data: bytes):
        logger = logging.getLogger(__name__)
        # 0-1: transid
        # 2: qr_flag[1] opcode[4] authoritative[1] trunc[1] rec_desired[1]
        # 3: rec_avail[1] zero[3] res_code[4]
        # 4-6: query_count
        # 7-8: answer_count
        # 9-10: authority_count
        # 11-12: additional_count
        transid, flags, qdcount, ancount, nscount, arcount = \
            struct.unpack_from('>6H', data) # raises ValueError
        logger.debug(f'{transid=} {flags=} {qdcount=} {ancount=} {nscount=} {arcount=}')
        return cls(
            transid = transid,
            isresp = bool(flags & (1 << 15)),
            opcode  = cls.Opcode((flags & (0xf << 11)) >> 11), # raises ValueError
            aa      = bool(flags & (1 << 10)),
            tc      = bool(flags & (1 << 9)),
            rd      = bool(flags & (1 << 8)),
            ra      = bool(flags & (1 << 7)),
            rcode   = cls.Rcode(flags & 0xf), # raises ValueError
            qdcount = qdcount,
            ancount = ancount,
            nscount = nscount,
            arcount = arcount
        )

@dataclass
class Question:
    qname: str
    qtype: int
    qclass: int

class RequestHandler:
    def __init__(self, request, client_address, server):
        self.request = request
        self.client_address = client_address
        self.server = server

    def handle2(self, req: bytes):
        logger = logging.getLogger('[%s]' % self.client_address[0])

        logger.debug(' '.join(hex(x)[2:].zfill(2) for x in req))
        req = BytesIO(req)

        try:
            header = Header.decode(req.read(12))
        except ValueError as e:
            logger.error(e)
            return
        logger.info(header)

        if header.isresp:
            logger.warning('Received thing is a response')
            return

        questions = []
        for i in range(header.qdcount):
            # 0-?: qname
            # +2: qtype
            # +2-+4: qclass
            qname = []
            try:
                while part_size := int.from_bytes(req.read(1)):
                    qname.append(req.read(part_size).decode('ascii')) # raises UnicodeDecodeError
            except UnicodeDecodeError as e:
                logger.error(e)
                return

            questions.append(Question(
                qname = '.'.join(qname),
                qtype = int.from_bytes(req.read(2)),
                qclass = int.from_bytes(req.read(2))
            ))
            logger.info(questions[-1])

    def handle_tcp(self):
        size = int.from_bytes(self.request.recv(2))
        req = self.request.recv(size)
        self.handle2(req)

    def handle_udp(self):
        req, sock = self.request
        self.handle2(req)

class Server:
    def __init__(self, server_address: (str, int), handler_class: RequestHandler):
        self.handler = handler_class
        self.executor = ThreadPoolExecutor()

        self.tcp = socket(AF_INET6, SOCK_STREAM)
        self.tcp.bind(server_address)
        self.tcp.listen()

        self.udp = socket(AF_INET6, SOCK_DGRAM)
        self.udp.bind(server_address)

        self.selector = DefaultSelector()
        self.selector.register(self.tcp, EVENT_READ)
        self.selector.register(self.udp, EVENT_READ)

    def serve_forever(self):
        while True:
            events = self.selector.select()
            for key, mask in events:
                if key.fileobj is self.tcp:
                    request, client_address = self.tcp.accept()
                    handler = self.handler(request, client_address, self)
                    handler.handle_tcp()
                elif key.fileobj is self.udp:
                    data, client_address = self.recvfrom(1024)
                    handler = self.handler(data, client_address, self)
                    handler.handle_udp()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        

if __name__ == '__main__':
    logging.basicConfig(
        level = logging.DEBUG,
        format = '%(asctime)s %(levelname)s %(name)s: %(message)s',
        datefmt = '%Y-%m-%d %H:%M:%S'
    )

    with Server(('', 253), RequestHandler) as server:
        try:
            server.serve_forever()
        except KeyboardInterrupt:
            server.server_close()
